#include <cmath>
#include <iostream>
#include <vector>

using namespace std;

#include "omp.h"
#include "mpi.h"

using integer = int;

int main(int argc, char **argv) 
{
  MPI_Init(&argc, &argv);

  int rank, size;
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &size);

  integer N = 10;
  integer M = 1;
  double debut, fin;
  double start, end;
  double time;

  debut = MPI_Wtime();

  for (int i = 0; i < argc; ++i) 
  {
    if (argv[i] == string("-n"))
      N = stol(argv[i + 1]);
    if (argv[i] == string("-m"))
      M = stol(argv[i + 1]);
  }
  double h = 0.5 * M_PI / (N - 1);

  integer Nloc = N / size;   // Local size for vectors
  integer rlimit = N % size; // Remainder of the euclidean division
  if (rank < rlimit)         // Case where N is not a multiple of size 
    Nloc++;

  if (rank == 0)
    cout << "Sizes: n= " << N << " , nloc= " << Nloc << " , m= " << M
         << " , h= " << h << endl;

  MPI_Barrier(MPI_COMM_WORLD);
  start = MPI_Wtime();

  vector<double> x(Nloc), y(Nloc);

  MPI_Barrier(MPI_COMM_WORLD);
  end = MPI_Wtime();
  time = end - start;
  if (rank == 0)
    cout << "allocation des vecteurs en " << temps << " s" << endl;

  MPI_Barrier(MPI_COMM_WORLD);
  start = MPI_Wtime();

  integer I;
  integer delta = 0; // Managing an offset for rlimit != 0.
  if (rank >= rlimit)
    delta = rlimit;

  #pragma omp parallel for
  for (integer i = 0; i < Nloc; ++i) 
  {
    I = i + rank * Nloc + delta;
    x[i] = pow(cos(I * h), 2);
    y[i] = pow(sin(I * h), 2);
  }

  MPI_Barrier(MPI_COMM_WORLD);
  end = MPI_Wtime();
  time = end - start;
  if (rank == 0)
    cout << "Vector initialized in " << time << " s" << endl;

  MPI_Barrier(MPI_COMM_WORLD);
  start = MPI_Wtime();

  double dotp_tot = 0.0;
  for (integer j = 0; j < M; ++j) 
  {
    double dotp,dotp_loc = 0;
    #pragma omp parallel for reduction(+ : dotp_loc)
    for (integer i = 0; i < Nloc; ++i)
      dotp_loc += x[i] * y[i];

    MPI_Allreduce(&dotp_loc, &dotp, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    dotp_tot += dotp;
  }
  MPI_Barrier(MPI_COMM_WORLD);
  end = MPI_Wtime();
  time = end - start;
  if (rank == 0)
    cout << "dotp_tot= " << dotp_tot << " computed in " << time << " s" << endl;

  fin = MPI_Wtime();
  time = fin - debut;
  if (rank == 0)
    cout << "Total time: " << time << " s" << endl;

  MPI_Finalize();
  return 0;
}
