#include <cmath>
#include <iostream>
#include <vector>
using namespace std;

#include "omp.h"

using integer = int;

int main(int argc, char **argv) 
{
  integer N = 10;
  integer M = 1;
  double debut, fin;
  double start, end;
  double time;

  debut = omp_get_wtime();

  for (int i = 0; i < argc; ++i) 
  {
    if (argv[i] == string("-n"))
      N = stol(argv[i + 1]);
    if (argv[i] == string("-m"))
      M = stol(argv[i + 1]);
  }
  double h = 0.5 * M_PI / (N - 1);
  cout << "Sizes: n= " << N << " ,  m= " << M << " , h= " << h << endl;

  start = omp_get_wtime();
  vector<double> x(N), y(N);
  end = omp_get_wtime();
  time = end - start;
  cout << "Vectors allocated in " << time << " s" << endl;

  start = omp_get_wtime();
  #pragma omp parallel for
  for (integer i = 0; i < N; ++i) 
  {
    x[i] = pow(cos(i * h), 2);
    y[i] = pow(sin(i * h), 2);
  }
  end = omp_get_wtime();
  time = end - start;
  cout << "Vector initialized in " << time << " s" << endl;

  start = omp_get_wtime();
  double dotp_tot = 0.0;
  for (integer j = 0; j < M; ++j) 
  {
    double dotp = 0.0;
    #pragma omp parallel for reduction(+ : dotp)
    for (integer i = 0; i < N; ++i)
      dotp += x[i] * y[i];

    dotp_tot += dotp;
  }
  end = omp_get_wtime();
  time = end - start;
  cout << "dotp_tot= " << dotp_tot << " computed in " << time << " s" << endl;

  fin = omp_get_wtime();
  time = fin - debut;
  cout << "Total time: " << time << " s" << endl;

  return 0;
}
