#include <cmath>
#include <ctime>
#include <iostream>
#include <string>
#include <vector>
#include "omp.h"

using namespace std;

int main(int argc, char **argv) 
{
  double a, b;

  a = omp_get_wtime();

  int n = 16;
  double epsilon = 1e-7;
  long max_iter = 1e8;

  for (int i = 0; i < argc; ++i) 
  {
    if (argv[i] == string("-n"))
      n = stoi(argv[i + 1]);
    if (argv[i] == string("-e"))
      epsilon = stod(argv[i + 1]);
    if (argv[i] == string("-i"))
      max_iter = stol(argv[i + 1]);
    if (argv[i] == string("-h")) {
      cout << "Usage : " << argv[0] << " [-n N=16] [-e E=1e-7] [-i I=1e8]"
           << endl;
      exit(0);
    }
  }

  cout << "command line : ";
  for (int i = 0; i < argc; ++i)
    cout << argv[i] << " ";
  cout << endl;

  int n2 = n + 2;
  int N = n2 * n2;
  double h = 1.0 / (n + 1);
  double h2 = h * h;

  cout << "N= " << n << " " << n2 << " " << N << " " << h << " " << h2 << endl;

  vector<double> f(N);
  #pragma omp parallel for
  for (int i = 0; i < N; i++)
    f[i] = 1.0;

  vector<double> x(N);
  #pragma omp parallel for
  for (int i = 0; i < N; i++)
    x[i] = 0.0;

  vector<double> oldx(N);

  long iter = 0;
  double error = 1.0;

  while (error > epsilon && iter < max_iter) 
  {
    #pragma omp parallel for
    for (int i = 0; i < N; i++)
      oldx[i] = x[i];

    #pragma omp parallel for
    for (int i = 1; i < n2 - 1; i++) 
    {
      for (int j = 1; j < n2 - 1; j++) 
      {
        int p = i * n2 + j;
        x[p] = 0.25 * (h2 * f[p] + (oldx[p + 1] + oldx[p - 1] + oldx[p + n2] +
                                    oldx[p - n2]));
      }
    }

    error = 0.0;
    #pragma omp parallel for reduction(+:error)
    for (int i = 0; i < N; i++)
      error += pow(x[i] - oldx[i], 2.0);
    error = sqrt(error);

    iter++;
  }
  b = omp_get_wtime();
  cout << "iter= " << iter << " error= " << error
       << " val= " << x[(n / 2 + 1) * n2 + n / 2 + 1] << " in "
       << (double)(b - a)  << " s" << endl;

  return 0;
}
